# -*- coding: utf-8 -*-
"""
Created on Mon Mar 11 08:28:46 2019

@author: Magnus
"""
#Initializing script and declearing packages
import numpy as np
import sys
import time
import pandas as pd
import pyomo.environ as pyo
from pyomo.opt import SolverFactory

# Declearing input-file and Data-dictionary to store input data
filename = "inputDeterministicmodel_21.xlsx"
Data = {}

#Declearing sets for generators and time-steps
generators = ["gen1", "gen2", "gen3", "gen4", "gen5"]
times = [1,2,3,4]

# Reading and storing generator data from excel sheet
df = pd.read_excel(filename, sheet_name = ("gendata"))
df = df.to_dict()
Data["gendata"] = df

"""
for gen in generators:
    print(df["NLC"][gen])
    #end for-loop
"""
# Reading and storing initial values for variable v
df = pd.read_excel(filename, sheet_name = ("init_v"))
df = df.to_dict()
Data["init_v"] = df

"""
print(Data["init_v"]["t1"]["gen3"])
"""
# Reading and storing data describing load, flexiramp requirement, etc in each timestep
df = pd.read_excel(filename, sheet_name = ("timedata"))
df = df.to_dict()
Data["timedata"] = df

# Reading and storing initial values for generator output
df = pd.read_excel(filename, sheet_name = ("init_g"))
df = df.to_dict()
Data["init_g"] = df

"""
CREATING PYOMO MODEL
"""

model = pyo.ConcreteModel()
"""
SETS
"""
# Set of the time-steps in the model
model.t = pyo.Set(initialize=times, ordered=True)
# Set containing the generators in the model
model.gen = pyo.Set(initialize=generators, ordered=True)

"""
PARAMETERS
"""
# Parameters conserning generators
# Maximum output of the generator in [MW]
model.Cap = pyo.Param(model.gen, initialize=Data["gendata"]["Cap"])
# Minimum output from generator while producing [MW]
model.CapMin = pyo.Param(model.gen, initialize=Data["gendata"]["CapMin"])
# Start-up cost of generator [$]
model.HC = pyo.Param(model.gen, initialize=Data["gendata"]["HC"])
# No-load cost (fixed cost) of running generator [$]
model.NLC = pyo.Param(model.gen, initialize=Data["gendata"]["NLC"])
# Operation cost (variable cost) of generator [$/MWh]
model.OC = pyo.Param(model.gen, initialize=Data["gendata"]["OC"])
# Ramp limit of generator [MW/interval]
model.RR = pyo.Param(model.gen, initialize=Data["gendata"]["RR"])
# Shut-down ramp limit of generator [MW/interval]
model.SDRR = pyo.Param(model.gen, initialize=Data["gendata"]["SDRR"])
# Start-up ramp limit of generator [MW/interval]
model.SURR = pyo.Param(model.gen, initialize=Data["gendata"]["SURR"])

# Parameters conserning timesteps
# Forecasted value of net load [MW]
model.DF = pyo.Param(model.t, initialize=Data["timedata"]["DF"])
# Forecast change in net load from one time period to the next [MW]
model.DNL = pyo.Param(model.t, initialize=Data["timedata"]["DNL"])
# Net load forecast error range [MW]
model.ERR = pyo.Param(model.t, initialize=Data["timedata"]["ERR"])
# Required down-flexiramp [MW/interval]
model.FRdn = pyo.Param(model.t, initialize=Data["timedata"]["FRdn"])
# Required up-flexiramp [MW/interval]
model.FRup = pyo.Param(model.t, initialize=Data["timedata"]["FRup"])
# Initial values for variables g and v
# Output of all generators in t=0 [MW]
model.Init_g = pyo.Param(model.t, initialize=Data["init_g"])
# Online status of all generators in t=0 and t=1
model.Init_v = pyo.Param(model.t, initialize=Data["init_v"])

"""
for t in ["t1","t2"] :
    for gen in generators :
        print(pyo.value(model.Init_v[t][gen]))
"""

# Parameters that are scalar
# Reference price for demand curve [$/MWh]
model.Pr_0 = pyo.Param(initialize=50)
# Slope of demand curve [($/MWh)/MW]
model.Slope = pyo.Param(initialize=-100)

"""
VARIABLES
"""
# Start-up cost of generator in time t
model.c  = pyo.Var(model.t, model.gen, within = pyo.NonNegativeReals)
# Ralized net load [MW] in time t. This equals the reference consumption of power, minus the sum of renewables and demand response
model.d = pyo.Var(model.t, within = pyo.NonNegativeReals)
# On/off status of generator i in time t. Binary variable.
model.v = pyo.Var(model.t, model.gen, within = pyo.Binary)
# Initializing v
for g in model.gen :
    for t in [1,2] :
        model.v[t,g]=model.Init_v[t][g]
        if t == 1 :
            model.v[t,g].fix()
# Down-flexiramp provided by generator [MW] in time t
model.dr = pyo.Var(model.t, model.gen, within = pyo.Reals)
# Up-flexiramp provided by generator [MW] in time t
model.ur = pyo.Var(model.t, model.gen, within = pyo.Reals)
# Generation of generating unit [MW] in time t 
model.g = pyo.Var(model.t, model.gen, within = pyo.NonNegativeReals)
# Initializing g !!! Fixing g in t=1 affects results !!!
for g in model.gen :
    model.g[1,g]=model.Init_g[1][g]
# Maximum feasible generation of genearator in time t [MW]
model.gMax = pyo.Var(model.t, model.gen, within = pyo.NonNegativeReals)


"""
OBJECTIVE FUNCTION
"""
# Objective function aiming to maximize social surplus
def Objective(model) :
    return(sum(0.25*(((model.Pr_0-model.Slope*model.DF[t])*model.d[t]+model.Slope*model.d[t]*model.d[t]/2) - (sum(model.OC[gen]*model.g[t,gen] for gen in model.gen)+sum(model.NLC[gen]*model.v[t,gen] for gen in model.gen)+sum(model.c[t,gen] for gen in model.gen))) for t in model.t))
model.OBJ   = pyo.Objective(rule = Objective, sense = pyo.maximize)

"""
CONSTRAINTS
"""
# Power balance
def PowerEq(model, t) :
    return(sum(model.g[t,gen] for gen in model.gen) - model.DF[t] == 0)
model.PE_const     = pyo.Constraint(model.t, rule = PowerEq)

# Up-flexiramp market clearing
def UpFlexClearing(model, t) :
    if t != 4 :
        return(sum(model.ur[t,gen] for gen in model.gen) >= model.FRup[t])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexClear   = pyo.Constraint(model.t, rule = UpFlexClearing)

# Down-Flexiramp market clearing
def DownFlexClearing(model, t) :
    if t != 4 :
        return(sum(model.dr[t,gen] for gen in model.gen) >= model.FRdn[t])
    else :
        return(pyo.Constraint.Skip)
model.DownFlexClear   = pyo.Constraint(model.t, rule = DownFlexClearing)

# Lower bound on Generation, Accounting for Start-up/shut-down
def GenBoundLo(model, t, gen):
    return(model.CapMin[gen]*model.v[t,gen] <= model.g[t,gen])
model.GenBoundLo    = pyo.Constraint(model.t, model.gen, rule = GenBoundLo)

# Upper bound on Generation
def GenBoundUp(model, t, gen):
    return(model.g[t,gen] <= model.gMax[t,gen])
model.GenBoundUp    = pyo.Constraint(model.t, model.gen, rule = GenBoundUp)

# Lower bound on generation - down-flexiramp. Accounting for start-up/shut-down
def GenFlexLo(model, t, gen):
    if t != 4 :
        return(model.CapMin[gen]*(model.v[t+1,gen]+model.v[t,gen]-1) <= -model.dr[t,gen]+model.g[t,gen])
    else :
        return(pyo.Constraint.Skip)
model.GenFlexLo     = pyo.Constraint(model.t, model.gen, rule = GenFlexLo)

# Upper bound on generation - down-flexiramp. Accounting for start-up/shut-down
def GenFlexUp(model, t, gen):
    if t != 4 :
        return(-model.dr[t,gen]+model.g[t,gen] <= model.gMax[t+1,gen]+model.Cap[gen]*(1-model.v[t+1,gen]))
    else :
        return(pyo.Constraint.Skip)
model.GenFlexUp     = pyo.Constraint(model.t, model.gen, rule = GenFlexUp)

# Lower bound on generation + up-flexiramp. Accounting for start-up/shut-down
def GenUpFlexLo(model, t, gen):
    if t != 4 :
        return(model.CapMin[gen]*(model.v[t+1,gen]+model.v[t,gen]-1) <= model.ur[t,gen] + model.g[t,gen])
    else :
        return(pyo.Constraint.Skip)
model.GenUpFlexLo     = pyo.Constraint(model.t, model.gen, rule = GenUpFlexLo)

# Upper bound on generation + up-flexiramp. Accounting for start-up/shut-down
def GenUpFlexUp(model, t, gen):
    if t != 4 :
        return(model.ur[t,gen] + model.g[t,gen] <= model.gMax[t+1,gen]+model.Cap[gen]*(1-model.v[t+1,gen]))
    else :
        return(pyo.Constraint.Skip)
model.GenUpFlexUp     = pyo.Constraint(model.t, model.gen, rule = GenUpFlexUp)

# Definition of Maximum feasible generation with regards to generator capacity
def MaxGenCap(model, t, gen) :
    return(model.gMax[t,gen] <= model.v[t,gen]*model.Cap[gen])
model.MaxGenCap     = pyo.Constraint(model.t, model.gen, rule = MaxGenCap)

# Definition of maximum feasible generation with regards to ramp rates and start-up RR
def FeasGenSU(model, t, gen) :
    if t != 1 :
        return(model.gMax[t,gen] <= model.g[t-1,gen] + model.RR[gen]*model.v[t-1,gen] + model.SURR[gen]*(model.v[t,gen]-model.v[t-1,gen]) + model.Cap[gen]*(1-model.v[t,gen]))
    else :
        return(pyo.Constraint.Skip)
model.FeasGenSU     = pyo.Constraint(model.t, model.gen, rule = FeasGenSU)

# Definition of maximum feasible generation with regards to ramp rates and shut down RR
def FeasGenSD(model, t, gen) :
    if t != 4 :
        return(model.gMax[t,gen] <= model.SDRR[gen]*(model.v[t,gen]-model.v[t+1,gen]) + model.Cap[gen]*model.v[t+1,gen])
    else :
        return(pyo.Constraint.Skip)
model.FeasGenSD     = pyo.Constraint(model.t, model.gen, rule = FeasGenSD)

# Ramp limits on Forecast Generation
def RampLimFore(model, t, gen) :
    if t != 1 :
        return(model.g[t-1,gen] - model.g[t,gen] <= model.RR[gen]*model.v[t,gen] + model.SDRR[gen]*(model.v[t-1,gen] - model.v[t,gen]) + model.Cap[gen]*(1-model.v[t-1,gen]))
    else :
        return(pyo.Constraint.Skip)
model.RampLimFore   = pyo.Constraint(model.t, model.gen, rule = RampLimFore)

# Lower bound on up-flexiramp with regards to Shut-Down
def UpFlexiBoundLo(model, t, gen) :
    if t != 4 :
        return(-model.RR[gen]*model.v[t+1,gen] - model.SDRR[gen]*(model.v[t,gen]-model.v[t+1,gen]) - model.Cap[gen]*(1-model.v[t,gen]) <= model.ur[t,gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiBoundLo     = pyo.Constraint(model.t, model.gen, rule = UpFlexiBoundLo)

# Upper bound on up-flexiramp with regards to start-up
def UpFlexiBoundHi(model, t, gen) :
    if t != 4 :
        return(model.ur[t,gen] <= model.RR[gen]*model.v[t,gen] + model.SURR[gen]*(model.v[t+1,gen]-model.v[t,gen]) + model.Cap[gen]*(1-model.v[t+1,gen]))
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiBoundHi     = pyo.Constraint(model.t, model.gen, rule = UpFlexiBoundHi)

# Lower bound on down-flexiramp with regards to start-up
def DnFlexiBoundLo(model, t, gen) :
    if t != 4 :
        return(-model.RR[gen]*model.v[t,gen] - model.SURR[gen]*(model.v[t+1,gen]-model.v[t,gen]) - model.Cap[gen]*(1-model.v[t+1,gen]) <= model.dr[t,gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiBoundLo     = pyo.Constraint(model.t, model.gen, rule = DnFlexiBoundLo)

# Upper bound on down-flexiramp with regards to shut-down
def DnFlexiBoundHi(model, t, gen) :
    if t != 4 :
        return(model.dr[t,gen] <= model.RR[gen]*model.v[t+1,gen] + model.SURR[gen]*(model.v[t,gen]-model.v[t+1,gen]) + model.Cap[gen]*(1-model.v[t,gen]))
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiBoundHi     = pyo.Constraint(model.t, model.gen, rule = DnFlexiBoundHi)

# Lower bound on up-flexiramp with regards to min/max operating capacity
def UpFlexiCapLo(model, t, gen) :
    if t != 4 :
        return(-model.v[t,gen]*model.Cap[gen] + model.v[t+1,gen]*model.CapMin[gen] <= model.ur[t,gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiCapLo      = pyo.Constraint(model.t, model.gen, rule = UpFlexiCapLo)

# Upper bound on up-flexiramp with regards to min/max operating capacity
def UpFlexiCapHi(model, t, gen) :
    if t != 4 :
        return(model.ur[t,gen] <= model.v[t+1,gen]*model.Cap[gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiCapHi      = pyo.Constraint(model.t, model.gen, rule = UpFlexiCapHi)

# Lower bound on down-flexiramp with regards to min/max operating capacity
def DnFlexiCapLo(model, t, gen) :
    if t != 4 :
        return(-model.v[t+1,gen]*model.Cap[gen] <= model.dr[t,gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiCapLo      = pyo.Constraint(model.t, model.gen, rule = DnFlexiCapLo)

# Upper bound on down-flexiramp with regards to min/max operating capacity
def DnFlexiCapHi(model, t, gen) :
    if t != 4 :
        return(model.dr[t,gen] <= model.v[t,gen]*model.Cap[gen] - model.v[t+1,gen]*model.CapMin[gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiCapHi      = pyo.Constraint(model.t, model.gen, rule = DnFlexiCapHi)

# Definition of start-up cost
def StartCost(model, t, gen) :
    if t != 1 :
        return(model.c[t,gen] >= model.HC[gen]*(model.v[t,gen]-model.v[t-1,gen]))
    else :
        return(pyo.Constraint.Skip)
model.StartCost     = pyo.Constraint(model.t, model.gen, rule = StartCost)

"""
SOLVING THE MODEL
"""
#Solver to use
opt = SolverFactory("gurobi")
    
#Allow finding dual data
model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

#Solve the problem
results = opt.solve(model, load_solutions = True)
    
#Display results
model.dual.display()
model.g.display()
model.dr.display()
model.ur.display()
## display all duals
#print ("Duals")
#for c in model.component_objects(pyo.Constraint, active=True):
#    print ("   Constraint",c)
#    for index in c:
#        print ("      ", index, model.dual[c[index]])
#print(model.dual[model.UpFlexClear[3]])


"""
STORING RESULTS
"""

# Resulting variables
Generation = {}
OnOffStatus = {}
Demand = {}
DownFlexiramp = {}
UpFlexiramp = {}
    
for t in model.t :
    Demand[t] = model.d[t].value
    Generation[t] = {}
    OnOffStatus[t] = {}
    DownFlexiramp[t] = {}
    UpFlexiramp[t] = {}
    for gen in model.gen :
        Generation[t][gen] = model.g[t,gen].value
        OnOffStatus[t][gen] = model.v[t,gen].value
        DownFlexiramp[t][gen] = model.dr[t,gen].value
        UpFlexiramp[t][gen] = model.ur[t,gen].value
            
#Convert to pandas
#We store each of the dictionary sections as a pandas Dataframe, so we can
#Store the data in an excel-file
        
Generation = pd.DataFrame(data=Generation)        #Store generation data
Demand = pd.DataFrame(data=Demand, index=[0])        #Store demand data
OnOffStatus = pd.DataFrame(data=OnOffStatus)        #Store On/off status of generators
DownFlexiramp = pd.DataFrame(data=DownFlexiramp)        #Store down-flexiramp data
UpFlexiramp = pd.DataFrame(data=UpFlexiramp)         #Store up-flexiramp data
        
#Then, we start storing this data into the output excel file
#Note that the excel file can exist before doing this procedure, if it doesn't it will be created
        

with pd.ExcelWriter("output_deterministic_21.xlsx") as writer:              #Set the location for excel file as the variable name 'writer'
           
    Generation.to_excel(writer, sheet_name= "Generation")      
    Demand.to_excel(writer, sheet_name= "Demand")       
    OnOffStatus.to_excel(writer, sheet_name= "OnOff status")         
    DownFlexiramp.to_excel(writer, sheet_name= "Down-flexiramp")        
    UpFlexiramp.to_excel(writer, sheet_name= "Up-flexiramp")        
