# -*- coding: utf-8 -*-
"""
Created on Mon May 20 14:46:34 2019

@author: Magnus
"""

#Initializing script and declearing packages
import numpy as np
import sys
import time
import pandas as pd
import pyomo.environ as pyo
from pyomo.opt import SolverFactory

# Declearing input-file and Data-dictionary to store input data
filename = "inputStochastic_23.xlsx"
Data = {}

#Declearing sets for generators and time-steps and scenarios
generators = ["gen1", "gen2", "gen3", "gen4", "gen5"]
times = [1,2,3,4]
df_realizations = [1, 2, 3]
Data["scen"] = {1: [1], 2: [1, 2, 3], 3: [1, 2, 3, 4, 5, 6, 7, 8, 9]}
templist = []
for i in range(1,28):
    templist.append(i)
Data["scen"][4] = templist

# Function for defining 2-dim Set
def funct(a,b):
    return((d,c) for d in a for c in b[d])


# Reading and storing generator data from excel sheet
df = pd.read_excel(filename, sheet_name = ("gendata"))
df = df.to_dict()
Data["gendata"] = df

"""
for gen in generators:
    print(df["NLC"][gen])
    #end for-loop
"""
# Reading and storing initial values for variable v
df = pd.read_excel(filename, sheet_name = ("init_v"))
df = df.to_dict()
Data["init_v"] = df

"""
print(Data["init_v"]["t1"]["gen3"])
"""
# Reading and storing data describing load, flexiramp requirement, etc in each timestep
df = pd.read_excel(filename, sheet_name = ("timedata"))
df = df.to_dict()
Data["timedata"] = df

# Reading and storing initial values for generator output
df = pd.read_excel(filename, sheet_name = ("init_g"))
df = df.to_dict()
Data["init_g"] = df

# Reading and storing initial values for scenario nodes
df = pd.read_excel(filename, sheet_name = ("df_scen"))
df = df.to_dict()
Data["df_scen"] = df

# Reading and storing initial values for probability of scenarios
df = pd.read_excel(filename, sheet_name = ("prob_scen"))
df = df.to_dict()
Data["prob_scen"] = df

# Reading and storing initial values for sequence of scenarios
df = pd.read_excel(filename, sheet_name = ("s_mark"))
df = df.to_dict()
Data["s_mark"] = df



"""
CREATING PYOMO MODEL
"""

model = pyo.ConcreteModel()
"""
SETS
"""
# Set of the time-steps in the model
model.t = pyo.Set(initialize=times, ordered=True)
# Set containing the generators in the model
model.gen = pyo.Set(initialize=generators, ordered=True)
# Set containing the scenarios in the model
model.scen = pyo.Set(dimen=2, initialize=funct(times, Data["scen"]), ordered=True)

model.scenset = pyo.Set(initialize=templist, ordered=True)

"""
PARAMETERS
"""
# Parameters conserning generators
# Maximum output of the generator in [MW]
model.Cap = pyo.Param(model.gen, initialize=Data["gendata"]["Cap"])
# Minimum output from generator while producing [MW]
model.CapMin = pyo.Param(model.gen, initialize=Data["gendata"]["CapMin"])
# Start-up cost of generator [$]
model.HC = pyo.Param(model.gen, initialize=Data["gendata"]["HC"])
# No-load cost (fixed cost) of running generator [$]
model.NLC = pyo.Param(model.gen, initialize=Data["gendata"]["NLC"])
# Operation cost (variable cost) of generator [$/MWh]
model.OC = pyo.Param(model.gen, initialize=Data["gendata"]["OC"])
# Ramp limit of generator [MW/interval]
model.RR = pyo.Param(model.gen, initialize=Data["gendata"]["RR"])
# Shut-down ramp limit of generator [MW/interval]
model.SDRR = pyo.Param(model.gen, initialize=Data["gendata"]["SDRR"])
# Start-up ramp limit of generator [MW/interval]
model.SURR = pyo.Param(model.gen, initialize=Data["gendata"]["SURR"])

# Parameters conserning timesteps
# Scenario values of net load [MW]    
model.DF = pyo.Param(model.scen, mutable=True)
for scen in model.scen:
    model.DF[scen] = Data["df_scen"][scen[0]][scen[1]]
# Forecast change in net load from one time period to the next [MW]
model.DNL = pyo.Param(model.t, initialize=Data["timedata"]["DNL"])
# Net load forecast error range [MW]
model.ERR = pyo.Param(model.t, initialize=Data["timedata"]["ERR"])
# Required down-flexiramp [MW/interval]
model.FRdn = pyo.Param(model.t, initialize=Data["timedata"]["FRdn"])
# Required up-flexiramp [MW/interval]
model.FRup = pyo.Param(model.t, initialize=Data["timedata"]["FRup"])
# Initial values for variables g and v
# Output of all generators in t=0 [MW]
model.Init_g = pyo.Param(model.t, initialize=Data["init_g"])
# Online status of all generators in t=0 and t=1
model.Init_v = pyo.Param(model.t, initialize=Data["init_v"])

# model.w = pyo.Param(model.t, initialize={1:27, 2:9, 3:3, 4:1})
model.s_mark = pyo.Param(model.scen, mutable=True)
for scen in model.scen:
    model.s_mark[scen] = int(Data["s_mark"][scen[0]][scen[1]])
    
# Parameter for probability of each scenario
model.Prob = pyo.Param(model.scen, mutable=True)
for scen in model.scen:
    model.Prob[scen] = Data["prob_scen"][scen[0]][scen[1]]

# Parameters that are scalar
# Reference price for demand curve [$/MWh]
model.Pr_0 = pyo.Param(initialize=50)
# Slope of demand curve [($/MWh)/MW]
model.Slope = pyo.Param(initialize=-100)

"""
VARIABLES
"""
# Start-up cost of generator in time t
model.c  = pyo.Var(model.scen, model.gen, within = pyo.NonNegativeReals)
# Ralized net load [MW] in time t. This equals the reference consumption of power, minus the sum of renewables and demand response
model.d = pyo.Var(model.scen, within = pyo.NonNegativeReals)
# On/off status of generator i in time t. Binary variable.
model.v = pyo.Var(model.scen, model.gen, within = pyo.Binary)
# Initializing v
for g in model.gen :
    for t in [1,2] :
        for s in [1,2,3] :
            if (t == 1) and (s == 1) :
                model.v[t,s,g]=model.Init_v[t][g]
                model.v[t,s,g].fix()
            elif t==1 and s>1 :
                pass
            else :
                model.v[t,s,g]=model.Init_v[t][g]
            
# Down-flexiramp provided by generator [MW] in time t
model.dr = pyo.Var(model.scen, model.gen, within = pyo.Reals)
# Up-flexiramp provided by generator [MW] in time t
model.ur = pyo.Var(model.scen, model.gen, within = pyo.Reals)
# Generation of generating unit [MW] in time t 
model.g = pyo.Var(model.scen, model.gen, within = pyo.NonNegativeReals)
# Initializing g !!! Fixing g in t=1 affects results !!!
for g in model.gen :
    model.g[1, 1, g]=model.Init_g[1][g]
# Maximum feasible generation of genearator in time t [MW]
model.gMax = pyo.Var(model.scen, model.gen, within = pyo.NonNegativeReals)


#for s in model.scen :
#    print(model.Prob[s])

# model.w[s[0]]*

"""
OBJECTIVE FUNCTION
"""
# Objective function aiming to maximize social surplus
def Objective(model) :
   return(sum(model.Prob[s]*0.25*(((model.Pr_0-model.Slope*model.DF[s])*model.d[s]+model.Slope*model.d[s]*model.d[s]/2) - (sum(model.OC[gen]*model.g[s, gen] for gen in model.gen)+sum(model.NLC[gen]*model.v[s,gen] for gen in model.gen)+sum(model.c[s,gen] for gen in model.gen))) for s in model.scen))
model.OBJ   = pyo.Objective(rule = Objective, sense = pyo.maximize)

"""
CONSTRAINTS
"""
# Power balance
def PowerEq(model, t, s) :
    return(sum(model.g[t,s,gen] for gen in model.gen) - model.DF[t,s] == 0)
model.PE_const     = pyo.Constraint(model.scen, rule = PowerEq)


# Up-flexiramp market clearing
def UpFlexClearing(model, t, s) :
    if t != 4 :
        return(sum(model.ur[t,s,gen] for gen in model.gen) >= model.FRup[t])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexClear   = pyo.Constraint(model.scen, rule = UpFlexClearing)

# Down-Flexiramp market clearing
def DownFlexClearing(model, t, s) :
    if t != 4 :
        return(sum(model.dr[t,s,gen] for gen in model.gen) >= model.FRdn[t])
    else :
        return(pyo.Constraint.Skip)
model.DownFlexClear   = pyo.Constraint(model.scen, rule = DownFlexClearing)

# Lower bound on Generation, Accounting for Start-up/shut-down
def GenBoundLo(model, t, s, gen):
    return(model.CapMin[gen]*model.v[t,s,gen] <= model.g[t,s,gen])
model.GenBoundLo    = pyo.Constraint(model.scen, model.gen, rule = GenBoundLo)

# Upper bound on Generation
def GenBoundUp(model, t, s, gen):
    return(model.g[t,s,gen] <= model.gMax[t,s,gen])
model.GenBoundUp    = pyo.Constraint(model.scen, model.gen, rule = GenBoundUp)

# Lower bound on generation - down-flexiramp. Accounting for start-up/shut-down
def GenFlexLo(model, t, s, gen):
    if t != 1 :
        return(model.CapMin[gen]*(model.v[t,s,gen]+model.v[(t-1),model.s_mark[t,s].value,gen]-1) <= -model.dr[t-1,model.s_mark[t,s].value,gen]+model.g[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.GenFlexLo     = pyo.Constraint(model.scen, model.gen, rule = GenFlexLo)


# Upper bound on generation - down-flexiramp. Accounting for start-up/shut-down
def GenFlexUp(model, t, s, gen):
    if t != 1 :
        return(-model.dr[t-1,model.s_mark[t,s].value,gen]+model.g[t-1,model.s_mark[t,s].value,gen] <= model.gMax[t,s,gen]+model.Cap[gen]*(1-model.v[t,s,gen]))
    else :
        return(pyo.Constraint.Skip)
model.GenFlexUp     = pyo.Constraint(model.scen, model.gen, rule = GenFlexUp)

# Lower bound on generation + up-flexiramp. Accounting for start-up/shut-down
def GenUpFlexLo(model, t, s, gen):
    if t != 1 :
        return(model.CapMin[gen]*(model.v[t,s,gen]+model.v[t-1,model.s_mark[t,s].value,gen]-1) <= model.ur[t-1,model.s_mark[t,s].value,gen] + model.g[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.GenUpFlexLo     = pyo.Constraint(model.scen, model.gen, rule = GenUpFlexLo)

# Upper bound on generation + up-flexiramp. Accounting for start-up/shut-down
def GenUpFlexUp(model, t, s, gen):
    if t != 1 :
        return(model.ur[t-1,model.s_mark[t,s].value,gen] + model.g[t-1,model.s_mark[t,s].value,gen] <= model.gMax[t,s,gen]+model.Cap[gen]*(1-model.v[t,s,gen]))
    else :
        return(pyo.Constraint.Skip)
model.GenUpFlexUp     = pyo.Constraint(model.scen, model.gen, rule = GenUpFlexUp)

# Definition of Maximum feasible generation with regards to generator capacity
def MaxGenCap(model, t, s, gen) :
    return(model.gMax[t,s,gen] <= model.v[t,s,gen]*model.Cap[gen])
model.MaxGenCap     = pyo.Constraint(model.scen, model.gen, rule = MaxGenCap)

# Definition of maximum feasible generation with regards to ramp rates and start-up RR
def FeasGenSU(model, t, s, gen) :
    if t != 1 :
        return(model.gMax[t, s, gen] <= model.g[t-1,model.s_mark[t,s].value,gen] + model.RR[gen]*model.v[t-1,model.s_mark[t,s].value,gen] + model.SURR[gen]*(model.v[t,s,gen]-model.v[t-1,model.s_mark[t,s].value,gen]) + model.Cap[gen]*(1-model.v[t,s,gen]))
    else :
        return(pyo.Constraint.Skip)
model.FeasGenSU     = pyo.Constraint(model.scen, model.gen, rule = FeasGenSU)

# Definition of maximum feasible generation with regards to ramp rates and shut down RR
def FeasGenSD(model, t, s, gen) :
    if t != 1 :
        return(model.gMax[t-1,model.s_mark[t,s].value,gen] <= model.SDRR[gen]*(model.v[t-1,model.s_mark[t,s].value,gen]-model.v[t,s,gen]) + model.Cap[gen]*model.v[t,s,gen])
    else :
        return(pyo.Constraint.Skip)
model.FeasGenSD     = pyo.Constraint(model.scen, model.gen, rule = FeasGenSD)

# Ramp limits on Forecast Generation
def RampLimFore(model, t, s, gen) :
    if t != 1 :
        return(model.g[t-1,model.s_mark[t,s].value,gen] - model.g[t,s,gen] <= model.RR[gen]*model.v[t,s,gen] + model.SDRR[gen]*(model.v[t-1,model.s_mark[t,s].value,gen] - model.v[t,s,gen]) + model.Cap[gen]*(1-model.v[t-1,model.s_mark[t,s].value,gen]))
    else :
        return(pyo.Constraint.Skip)
model.RampLimFore   = pyo.Constraint(model.scen, model.gen, rule = RampLimFore)

# Lower bound on up-flexiramp with regards to Shut-Down
def UpFlexiBoundLo(model, t, s, gen) :
    if t != 1 :
        return(-model.RR[gen]*model.v[t,s,gen] - model.SDRR[gen]*(model.v[t-1,model.s_mark[t,s].value,gen]-model.v[t,s,gen]) - model.Cap[gen]*(1-model.v[t-1,model.s_mark[t,s].value,gen]) <= model.ur[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiBoundLo     = pyo.Constraint(model.scen, model.gen, rule = UpFlexiBoundLo)

# Upper bound on up-flexiramp with regards to start-up
def UpFlexiBoundHi(model, t, s, gen) :
    if t != 1 :
        return(model.ur[t-1,model.s_mark[t,s].value,gen] <= model.RR[gen]*model.v[t-1,model.s_mark[t,s].value,gen] + model.SURR[gen]*(model.v[t,s,gen]-model.v[t-1,model.s_mark[t,s].value,gen]) + model.Cap[gen]*(1-model.v[t,s,gen]))
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiBoundHi     = pyo.Constraint(model.scen, model.gen, rule = UpFlexiBoundHi)

# Lower bound on down-flexiramp with regards to start-up
def DnFlexiBoundLo(model, t, s, gen) :
    if t != 1 :
        return(-model.RR[gen]*model.v[t-1,model.s_mark[t,s].value,gen] - model.SURR[gen]*(model.v[t,s,gen]-model.v[t-1,model.s_mark[t,s].value,gen]) - model.Cap[gen]*(1-model.v[t,s,gen]) <= model.dr[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiBoundLo     = pyo.Constraint(model.scen, model.gen, rule = DnFlexiBoundLo)

# Upper bound on down-flexiramp with regards to shut-down
def DnFlexiBoundHi(model, t, s, gen) :
    if t != 1 :
        return(model.dr[t-1,model.s_mark[t,s].value,gen] <= model.RR[gen]*model.v[t,s,gen] + model.SURR[gen]*(model.v[t-1,model.s_mark[t,s].value,gen]-model.v[t,s,gen]) + model.Cap[gen]*(1-model.v[t-1,model.s_mark[t,s].value,gen]))
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiBoundHi     = pyo.Constraint(model.scen, model.gen, rule = DnFlexiBoundHi)

# Lower bound on up-flexiramp with regards to min/max operating capacity
def UpFlexiCapLo(model, t, s, gen) :
    if t != 1 :
        return(-model.v[t-1,model.s_mark[t,s].value,gen]*model.Cap[gen] + model.v[t,model.s_mark[t,s].value,gen]*model.CapMin[gen] <= model.ur[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiCapLo      = pyo.Constraint(model.scen, model.gen, rule = UpFlexiCapLo)

# Upper bound on up-flexiramp with regards to min/max operating capacity
def UpFlexiCapHi(model, t, s, gen) :
    if t != 1 :
        return(model.ur[t-1,model.s_mark[t,s].value,gen] <= model.v[t,s,gen]*model.Cap[gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiCapHi      = pyo.Constraint(model.scen, model.gen, rule = UpFlexiCapHi)

# Lower bound on down-flexiramp with regards to min/max operating capacity
def DnFlexiCapLo(model, t, s, gen) :
    if t != 1 :
        return(-model.v[t,s,gen]*model.Cap[gen] <= model.dr[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiCapLo      = pyo.Constraint(model.scen, model.gen, rule = DnFlexiCapLo)

# Upper bound on down-flexiramp with regards to min/max operating capacity
def DnFlexiCapHi(model, t, s, gen) :
    if t != 1 :
        return(model.dr[t-1,model.s_mark[t,s].value,gen] <= model.v[t-1,model.s_mark[t,s].value,gen]*model.Cap[gen] - model.v[t,s,gen]*model.CapMin[gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiCapHi      = pyo.Constraint(model.scen, model.gen, rule = DnFlexiCapHi)

# Definition of start-up cost
def StartCost(model, t, s, gen) :
    if t != 1 :
        return(model.c[t,s,gen] >= model.HC[gen]*(model.v[t,s,gen]-model.v[t-1,model.s_mark[t,s].value,gen]))
    else :
        return(pyo.Constraint.Skip)
model.StartCost     = pyo.Constraint(model.scen, model.gen, rule = StartCost)

"""
SOLVING THE MODEL
"""
#Solver to use
opt = SolverFactory("gurobi")
    
#Allow finding dual data
model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

#Solve the problem
results = opt.solve(model, load_solutions = True)
    
#Display results
model.dual.display()
#Display results
model.OBJ.display()

  
ObjScen = sum(0.25*(((model.Pr_0-model.Slope*model.DF[s].value)*model.d[s].value+model.Slope*model.d[s].value*model.d[s].value/2) - (sum(model.OC[gen]*model.g[s, gen].value for gen in model.gen)+sum(model.NLC[gen]*model.v[s,gen].value for gen in model.gen)+sum(model.c[s,gen].value for gen in model.gen))) for s in [(1,1),(2,2),(3,5),(4,14)])
print('Objective value for basis scen is:' , ObjScen)
"""
SECOND RUN TO DETERMINE DUALS
"""
OnOffStatus = {}
# Extracting final values for v
for scen in model.scen :
    OnOffStatus[scen] = {}
    for gen in model.gen :
        OnOffStatus[scen][gen] = model.v[scen,gen].value
"""
CREATING PYOMO MODEL
"""

model = pyo.ConcreteModel()
"""
SETS
"""
# Set of the time-steps in the model
model.t = pyo.Set(initialize=times, ordered=True)
# Set containing the generators in the model
model.gen = pyo.Set(initialize=generators, ordered=True)
# Set containing the scenarios in the model
model.scen = pyo.Set(dimen=2, initialize=funct(times, Data["scen"]), ordered=True)

model.scenset = pyo.Set(initialize=templist, ordered=True)

"""
PARAMETERS
"""
# Parameters conserning generators
# Maximum output of the generator in [MW]
model.Cap = pyo.Param(model.gen, initialize=Data["gendata"]["Cap"])
# Minimum output from generator while producing [MW]
model.CapMin = pyo.Param(model.gen, initialize=Data["gendata"]["CapMin"])
# Start-up cost of generator [$]
model.HC = pyo.Param(model.gen, initialize=Data["gendata"]["HC"])
# No-load cost (fixed cost) of running generator [$]
model.NLC = pyo.Param(model.gen, initialize=Data["gendata"]["NLC"])
# Operation cost (variable cost) of generator [$/MWh]
model.OC = pyo.Param(model.gen, initialize=Data["gendata"]["OC"])
# Ramp limit of generator [MW/interval]
model.RR = pyo.Param(model.gen, initialize=Data["gendata"]["RR"])
# Shut-down ramp limit of generator [MW/interval]
model.SDRR = pyo.Param(model.gen, initialize=Data["gendata"]["SDRR"])
# Start-up ramp limit of generator [MW/interval]
model.SURR = pyo.Param(model.gen, initialize=Data["gendata"]["SURR"])

# Parameters conserning timesteps
# Scenario values of net load [MW]    
model.DF = pyo.Param(model.scen, mutable=True)
for scen in model.scen:
    model.DF[scen] = Data["df_scen"][scen[0]][scen[1]]
# Forecast change in net load from one time period to the next [MW]
model.DNL = pyo.Param(model.t, initialize=Data["timedata"]["DNL"])
# Net load forecast error range [MW]
model.ERR = pyo.Param(model.t, initialize=Data["timedata"]["ERR"])
# Required down-flexiramp [MW/interval]
model.FRdn = pyo.Param(model.t, initialize=Data["timedata"]["FRdn"])
# Required up-flexiramp [MW/interval]
model.FRup = pyo.Param(model.t, initialize=Data["timedata"]["FRup"])
# Initial values for variables g and v
# Output of all generators in t=0 [MW]
model.Init_g = pyo.Param(model.t, initialize=Data["init_g"])
# Online status of all generators in t=0 and t=1
model.Init_v = pyo.Param(model.t, initialize=Data["init_v"])

# model.w = pyo.Param(model.t, initialize={1:27, 2:9, 3:3, 4:1})
model.s_mark = pyo.Param(model.scen, mutable=True)
for scen in model.scen:
    model.s_mark[scen] = int(Data["s_mark"][scen[0]][scen[1]])
    
# Parameter for probability of each scenario
model.Prob = pyo.Param(model.scen, mutable=True)
for scen in model.scen:
    model.Prob[scen] = Data["prob_scen"][scen[0]][scen[1]]

# Parameters that are scalar
# Reference price for demand curve [$/MWh]
model.Pr_0 = pyo.Param(initialize=50)
# Slope of demand curve [($/MWh)/MW]
model.Slope = pyo.Param(initialize=-100)

"""
VARIABLES
"""
# Start-up cost of generator in time t
model.c  = pyo.Var(model.scen, model.gen, within = pyo.NonNegativeReals)
# Ralized net load [MW] in time t. This equals the reference consumption of power, minus the sum of renewables and demand response
model.d = pyo.Var(model.scen, within = pyo.NonNegativeReals)
# On/off status of generator i in time t. Binary variable.
model.v = pyo.Param(model.scen, model.gen, mutable=True)
# Initializing v
for scen in model.scen :
    for g in model.gen :
        model.v[scen,g] = OnOffStatus[scen][g]
            
# Down-flexiramp provided by generator [MW] in time t
model.dr = pyo.Var(model.scen, model.gen, within = pyo.Reals)
# Up-flexiramp provided by generator [MW] in time t
model.ur = pyo.Var(model.scen, model.gen, within = pyo.Reals)
# Generation of generating unit [MW] in time t 
model.g = pyo.Var(model.scen, model.gen, within = pyo.NonNegativeReals)
# Initializing g !!! Fixing g in t=1 affects results !!!
for g in model.gen :
    model.g[1, 1, g]=model.Init_g[1][g]
# Maximum feasible generation of genearator in time t [MW]
model.gMax = pyo.Var(model.scen, model.gen, within = pyo.NonNegativeReals)


#for s in model.scen :
#    print(model.Prob[s])

# model.w[s[0]]*

"""
OBJECTIVE FUNCTION
"""
# Objective function aiming to maximize social surplus
def Objective(model) :
   return(sum(model.Prob[s]*0.25*(((model.Pr_0-model.Slope*model.DF[s])*model.d[s]+model.Slope*model.d[s]*model.d[s]/2) - (sum(model.OC[gen]*model.g[s, gen] for gen in model.gen)+sum(model.NLC[gen]*model.v[s,gen] for gen in model.gen)+sum(model.c[s,gen] for gen in model.gen))) for s in model.scen))
model.OBJ   = pyo.Objective(rule = Objective, sense = pyo.maximize)

"""
CONSTRAINTS
"""
# Power balance
def PowerEq(model, t, s) :
    return(sum(model.g[t,s,gen] for gen in model.gen) - model.DF[t,s] == 0)
model.PE_const     = pyo.Constraint(model.scen, rule = PowerEq)


# Up-flexiramp market clearing
def UpFlexClearing(model, t, s) :
    if t != 4 :
        return(sum(model.ur[t,s,gen] for gen in model.gen) >= model.FRup[t])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexClear   = pyo.Constraint(model.scen, rule = UpFlexClearing)

# Down-Flexiramp market clearing
def DownFlexClearing(model, t, s) :
    if t != 4 :
        return(sum(model.dr[t,s,gen] for gen in model.gen) >= model.FRdn[t])
    else :
        return(pyo.Constraint.Skip)
model.DownFlexClear   = pyo.Constraint(model.scen, rule = DownFlexClearing)

# Lower bound on Generation, Accounting for Start-up/shut-down
def GenBoundLo(model, t, s, gen):
    return(model.CapMin[gen]*model.v[t,s,gen] <= model.g[t,s,gen])
model.GenBoundLo    = pyo.Constraint(model.scen, model.gen, rule = GenBoundLo)

# Upper bound on Generation
def GenBoundUp(model, t, s, gen):
    return(model.g[t,s,gen] <= model.gMax[t,s,gen])
model.GenBoundUp    = pyo.Constraint(model.scen, model.gen, rule = GenBoundUp)

# Lower bound on generation - down-flexiramp. Accounting for start-up/shut-down
def GenFlexLo(model, t, s, gen):
    if t != 1 :
        return(model.CapMin[gen]*(model.v[t,s,gen]+model.v[(t-1),model.s_mark[t,s].value,gen]-1) <= -model.dr[t-1,model.s_mark[t,s].value,gen]+model.g[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.GenFlexLo     = pyo.Constraint(model.scen, model.gen, rule = GenFlexLo)


# Upper bound on generation - down-flexiramp. Accounting for start-up/shut-down
def GenFlexUp(model, t, s, gen):
    if t != 1 :
        return(-model.dr[t-1,model.s_mark[t,s].value,gen]+model.g[t-1,model.s_mark[t,s].value,gen] <= model.gMax[t,s,gen]+model.Cap[gen]*(1-model.v[t,s,gen]))
    else :
        return(pyo.Constraint.Skip)
model.GenFlexUp     = pyo.Constraint(model.scen, model.gen, rule = GenFlexUp)

# Lower bound on generation + up-flexiramp. Accounting for start-up/shut-down
def GenUpFlexLo(model, t, s, gen):
    if t != 1 :
        return(model.CapMin[gen]*(model.v[t,s,gen]+model.v[t-1,model.s_mark[t,s].value,gen]-1) <= model.ur[t-1,model.s_mark[t,s].value,gen] + model.g[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.GenUpFlexLo     = pyo.Constraint(model.scen, model.gen, rule = GenUpFlexLo)

# Upper bound on generation + up-flexiramp. Accounting for start-up/shut-down
def GenUpFlexUp(model, t, s, gen):
    if t != 1 :
        return(model.ur[t-1,model.s_mark[t,s].value,gen] + model.g[t-1,model.s_mark[t,s].value,gen] <= model.gMax[t,s,gen]+model.Cap[gen]*(1-model.v[t,s,gen]))
    else :
        return(pyo.Constraint.Skip)
model.GenUpFlexUp     = pyo.Constraint(model.scen, model.gen, rule = GenUpFlexUp)

# Definition of Maximum feasible generation with regards to generator capacity
def MaxGenCap(model, t, s, gen) :
    return(model.gMax[t,s,gen] <= model.v[t,s,gen]*model.Cap[gen])
model.MaxGenCap     = pyo.Constraint(model.scen, model.gen, rule = MaxGenCap)

# Definition of maximum feasible generation with regards to ramp rates and start-up RR
def FeasGenSU(model, t, s, gen) :
    if t != 1 :
        return(model.gMax[t, s, gen] <= model.g[t-1,model.s_mark[t,s].value,gen] + model.RR[gen]*model.v[t-1,model.s_mark[t,s].value,gen] + model.SURR[gen]*(model.v[t,s,gen]-model.v[t-1,model.s_mark[t,s].value,gen]) + model.Cap[gen]*(1-model.v[t,s,gen]))
    else :
        return(pyo.Constraint.Skip)
model.FeasGenSU     = pyo.Constraint(model.scen, model.gen, rule = FeasGenSU)

# Definition of maximum feasible generation with regards to ramp rates and shut down RR
def FeasGenSD(model, t, s, gen) :
    if t != 1 :
        return(model.gMax[t-1,model.s_mark[t,s].value,gen] <= model.SDRR[gen]*(model.v[t-1,model.s_mark[t,s].value,gen]-model.v[t,s,gen]) + model.Cap[gen]*model.v[t,s,gen])
    else :
        return(pyo.Constraint.Skip)
model.FeasGenSD     = pyo.Constraint(model.scen, model.gen, rule = FeasGenSD)

# Ramp limits on Forecast Generation
def RampLimFore(model, t, s, gen) :
    if t != 1 :
        return(model.g[t-1,model.s_mark[t,s].value,gen] - model.g[t,s,gen] <= model.RR[gen]*model.v[t,s,gen] + model.SDRR[gen]*(model.v[t-1,model.s_mark[t,s].value,gen] - model.v[t,s,gen]) + model.Cap[gen]*(1-model.v[t-1,model.s_mark[t,s].value,gen]))
    else :
        return(pyo.Constraint.Skip)
model.RampLimFore   = pyo.Constraint(model.scen, model.gen, rule = RampLimFore)

# Lower bound on up-flexiramp with regards to Shut-Down
def UpFlexiBoundLo(model, t, s, gen) :
    if t != 1 :
        return(-model.RR[gen]*model.v[t,s,gen] - model.SDRR[gen]*(model.v[t-1,model.s_mark[t,s].value,gen]-model.v[t,s,gen]) - model.Cap[gen]*(1-model.v[t-1,model.s_mark[t,s].value,gen]) <= model.ur[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiBoundLo     = pyo.Constraint(model.scen, model.gen, rule = UpFlexiBoundLo)

# Upper bound on up-flexiramp with regards to start-up
def UpFlexiBoundHi(model, t, s, gen) :
    if t != 1 :
        return(model.ur[t-1,model.s_mark[t,s].value,gen] <= model.RR[gen]*model.v[t-1,model.s_mark[t,s].value,gen] + model.SURR[gen]*(model.v[t,s,gen]-model.v[t-1,model.s_mark[t,s].value,gen]) + model.Cap[gen]*(1-model.v[t,s,gen]))
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiBoundHi     = pyo.Constraint(model.scen, model.gen, rule = UpFlexiBoundHi)

# Lower bound on down-flexiramp with regards to start-up
def DnFlexiBoundLo(model, t, s, gen) :
    if t != 1 :
        return(-model.RR[gen]*model.v[t-1,model.s_mark[t,s].value,gen] - model.SURR[gen]*(model.v[t,s,gen]-model.v[t-1,model.s_mark[t,s].value,gen]) - model.Cap[gen]*(1-model.v[t,s,gen]) <= model.dr[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiBoundLo     = pyo.Constraint(model.scen, model.gen, rule = DnFlexiBoundLo)

# Upper bound on down-flexiramp with regards to shut-down
def DnFlexiBoundHi(model, t, s, gen) :
    if t != 1 :
        return(model.dr[t-1,model.s_mark[t,s].value,gen] <= model.RR[gen]*model.v[t,s,gen] + model.SURR[gen]*(model.v[t-1,model.s_mark[t,s].value,gen]-model.v[t,s,gen]) + model.Cap[gen]*(1-model.v[t-1,model.s_mark[t,s].value,gen]))
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiBoundHi     = pyo.Constraint(model.scen, model.gen, rule = DnFlexiBoundHi)

# Lower bound on up-flexiramp with regards to min/max operating capacity
def UpFlexiCapLo(model, t, s, gen) :
    if t != 1 :
        return(-model.v[t-1,model.s_mark[t,s].value,gen]*model.Cap[gen] + model.v[t,model.s_mark[t,s].value,gen]*model.CapMin[gen] <= model.ur[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiCapLo      = pyo.Constraint(model.scen, model.gen, rule = UpFlexiCapLo)

# Upper bound on up-flexiramp with regards to min/max operating capacity
def UpFlexiCapHi(model, t, s, gen) :
    if t != 1 :
        return(model.ur[t-1,model.s_mark[t,s].value,gen] <= model.v[t,s,gen]*model.Cap[gen])
    else :
        return(pyo.Constraint.Skip)
model.UpFlexiCapHi      = pyo.Constraint(model.scen, model.gen, rule = UpFlexiCapHi)

# Lower bound on down-flexiramp with regards to min/max operating capacity
def DnFlexiCapLo(model, t, s, gen) :
    if t != 1 :
        return(-model.v[t,s,gen]*model.Cap[gen] <= model.dr[t-1,model.s_mark[t,s].value,gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiCapLo      = pyo.Constraint(model.scen, model.gen, rule = DnFlexiCapLo)

# Upper bound on down-flexiramp with regards to min/max operating capacity
def DnFlexiCapHi(model, t, s, gen) :
    if t != 1 :
        return(model.dr[t-1,model.s_mark[t,s].value,gen] <= model.v[t-1,model.s_mark[t,s].value,gen]*model.Cap[gen] - model.v[t,s,gen]*model.CapMin[gen])
    else :
        return(pyo.Constraint.Skip)
model.DnFlexiCapHi      = pyo.Constraint(model.scen, model.gen, rule = DnFlexiCapHi)

# Definition of start-up cost
def StartCost(model, t, s, gen) :
    if t != 1 :
        return(model.c[t,s,gen] >= model.HC[gen]*(model.v[t,s,gen]-model.v[t-1,model.s_mark[t,s].value,gen]))
    else :
        return(pyo.Constraint.Skip)
model.StartCost     = pyo.Constraint(model.scen, model.gen, rule = StartCost)

"""
SOLVING THE MODEL
"""
#Solver to use
opt = SolverFactory("gurobi")
    
#Allow finding dual data
model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

#Solve the problem
results = opt.solve(model, load_solutions = True)
    
#Display results
#model.dual.display()


market_price = np.array([model.dual[model.PE_const[scen]] for scen in [(1,1),(2,2),(3,5),(4,14)]])
upFlex_price = np.array([model.dual[model.UpFlexClear[scen]] for scen in [(1,1),(2,2),(3,5)]])
dnFlex_price = np.array([model.dual[model.DownFlexClear[scen]] for scen in [(1,1),(2,2),(3,5)]])
print(market_price)
print(upFlex_price)
print(dnFlex_price)